/*	File			Game_Server_ServerAdvertise_ClientDisconnect.java
	Purpose			Network Game (Noughts and Crosses)SERVER
	Author			Richard John Anthony
	Date			February 2015 (original Cpp version December 2003)
*/
package Game_Server_ServerAdvertise_ClientDisconnect;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.*;
import java.util.Arrays;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Random;
import javax.swing.ScrollPaneConstants;
import javax.swing.JOptionPane;

public class Game_Server_ServerAdvertise_ClientDisconnect extends javax.swing.JFrame implements ActionListener
{
    private static int PORT = 8000;                 // Default port 8000 decimal (clients connect to the server on this port)
    private static int LOCAL_PORT = 8008;           // Port that ServerAdvertise socket binds to (so not tying up the SERVER_ADVERTISE_PORT,
                                                    // thus allowing a Game client to be co-resident on the same computer as the Server)
    private static int SERVER_ADVERTISE_PORT = 8009;// Server Advertisement port 8009 decimal (clients listen on this port)
    private static int MAX_ALIAS_LEN = 15;
    private static int MAX_SERVER_NAME_LEN = 30;

    public enum MessageType {REGISTER_ALIAS, PLAYER_LIST, CHOOSE_OPPONENT, LOCAL_CELL_SELECTION,
            OPPONENT_CELL_SELECTION, START_OF_GAME, END_OF_GAME, CLIENT_DISCONNECT, SERVER_ADVERTISE};

    public enum CodeType {END_OF_GAME_DRAW, END_OF_GAME_WIN, END_OF_GAME_LOSE, END_OF_GAME_CLIENT_DISCONNECT};

    // To implement the equivallent of a C/C++/C# structure, in Java a class is used
    final class Game_Message_PDU
    {
        MessageType eMessageType;   //size 4                Offset 0
        byte[] cAlias;              //size [15+1]           Offset 4
        byte[] cPlayerList;         //size [10 * (15+1)]    Offset 20   Comma-separated list of aliases
        int iCell;                  //size 4                Offset 180
        CodeType eCode;             //size 4                Offset 184
        int iToken;                 //size 4                Offset 188
        Boolean bFirstMove;         //size 1(padded to 4)   Offset 192
    };

    private static int Game_Message_PDU_Size = 196;
        
    final class ServerAdvertise_Message_PDU
    {
        MessageType eMessageType;   //size 4        Offset 0
        byte[] cServerName;         //size [30+1]   Offest 4
        byte iIP_Address_Byte_1;    //size 1        Offset 35
        byte iIP_Address_Byte_2;    //size 1        Offset 36
        byte iIP_Address_Byte_3;    //size 1        Offset 37
        byte iIP_Address_Byte_4;    //size 1        Offset 38
        // Note - Field alignments (compatibility with C++ version of the game) - Due to the nature of the serialisation
        // performed in the C++ version, all 4-byte fields are aligned on 4-byte-boundaries. The cServerName field is 
        // 31 bytes long, and thus ends not on a 4-byte boundary. The four single byte fields that follow it are aligned
        // on byte boundaries. This means that in the actual transmitted byte array, a single padding byte (value 0) is
        // inserted before the iPort field, this requires attention when serialising / deserialising in the C# versions 
        // of the application, see DeSerialize_ReceivedServerAdvertisementMessage()
        // Port values are actually Int16, but encoded as Int32 in this message format for compatibility with the pre-existing C++ version
        int iPort;                  //size 4        Offset 40
    };
     
    private static int ServerAdvertise_Message_PDU_Size = 44;
    byte[] m_bSendServerAdvertisementBuffer; // A serialised byte buffer representation of ServerAdvertise_Message_PDU
    
    ServerSocket m_ListenSocket;
    DatagramSocket m_SendServerAdvertisementSocket;
    int m_iPort;
    InetAddress  m_LocalAddress;
    InetAddress  m_SendServerAdvertisementAddress;
    Timer m_Receive_Timer;
    Timer m_SendServerAdvertisement_Timer;
    
    private static int MAX_GAMES = 5;
    private static int MAX_CONNECTIONS = MAX_GAMES * 2;

    final class Connection
    {
        public Socket ConnectSock;
        InputStream InStream;
        OutputStream OutStream;
        public String sAlias;
        public Boolean bInUse;
    };

    final class Game
    {
        public int[] iConnection;   // Two connections (two clients) per game
        public Boolean bInUse;
        public int[][] iGrid;        // 3 x 3 noughts and crosses game grid
    };
        
    Connection[] m_ConnectionArray;
    Game[] m_GameArray;

    Random m_Random; // Needed for determining which player has first turn in a newly-created game

    public Game_Server_ServerAdvertise_ClientDisconnect()
    {
        initComponents();
        m_ConnectionArray = new Connection[MAX_CONNECTIONS];
        m_GameArray = new Game[MAX_GAMES];
        m_Random = new Random(); // Automatically randomly seed Random Number Generator
        m_iPort = PORT;
        m_LocalAddress = GetLocalIPAddress();
        
        for(int i = 0; i < MAX_CONNECTIONS; i++)
        {	
            m_ConnectionArray[i] = new Connection();
            m_ConnectionArray[i].bInUse = false;
            m_ConnectionArray[i].sAlias = "";
        }	
        for(int i = 0; i < MAX_GAMES; i++)
        {	
            m_GameArray[i] = new Game();
            m_GameArray[i].iConnection = new int[2];
            m_GameArray[i].bInUse = false;
            m_GameArray[i].iGrid = new int[3][3];
        }

        jTextField_Port.setText(Integer.toString(m_iPort));
        jTextField_Port.setEnabled(true);
        jTextField_IPAddress.setEnabled(false);
        jButton_EnableServer.setEnabled(true);
        jTextField_NumActiveGames.setEnabled(false);
        jTextField_NumConnectedClients.setEnabled(false);
        CountActiveConnections();   // Updates control
        CountActiveGames();         // Updates control
    }

    public InetAddress GetLocalIPAddress()
    {   
        String sHostAddress = "";
        try
        {
            InetAddress LocalAddress = InetAddress.getLocalHost();
            sHostAddress = LocalAddress.toString(); // Expected format: "Hostname/192.168.100.5"
            String sStr = "";
            String sHostName = "";
            int iStartOfDottedIPSubstring = sHostAddress.indexOf('/');
            if(-1 == iStartOfDottedIPSubstring)
            {   // Align to the actual start of the dotted IP address part of the string
                iStartOfDottedIPSubstring = 0; // '/' not found
            }
            else
            {
                sHostName = sHostAddress.substring(0, iStartOfDottedIPSubstring);
                iStartOfDottedIPSubstring++; // Skip the '/' charachter
            }
            String sDottedIPAddress = sHostAddress.substring(iStartOfDottedIPSubstring);
            sStr = String.format("Local host:%s,  IPaddress:%s", sHostName, sDottedIPAddress);
            WriteEventLogEntry(sStr);
            jTextField_IPAddress.setText(sDottedIPAddress);
            jButton_EnableServer.setEnabled(true);
            return LocalAddress;
        }
        catch (Exception Ex)
        {
            JOptionPane.showMessageDialog(null,Ex.toString(),"NoughtsCrossesSERVER",0);
            CloseAndExit();
        }
        return null;
    }

    void CreateSocket() // Called from 'actionPerformed' for JButton_EnableServer
    {
        // The Port number can be overwritten by the user since it was automatically
        // set to the default value. Hence the need to read the final value
        m_iPort = Integer.parseInt(jTextField_Port.getText());

        try
        {
            m_ListenSocket = new ServerSocket(m_iPort, 5/*backlog*/, InetAddress.getLocalHost()/*bindAddr*/);
            m_ListenSocket.setSoTimeout(100); // Timeout = 100mS i.e. non-blocking
        }
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),"NoughtsCrossesSERVER",0);
           CloseAndExit();
        }
    }

    void Accept()
    {
        String sStr = "";
        int iIndex = GetIndexOfUnusedConnection();
        if (iIndex == -1)
        {
            WriteEventLogEntry("Accept refused - Maximum number of Connections reached");
            return;
        }
        try
        {
            m_ConnectionArray[iIndex].ConnectSock = m_ListenSocket.accept();
        }
        catch (SocketTimeoutException Ex)
        {   
            return;
        }
        catch (Exception Ex)
        {
            sStr = String.format("Accept failed, %s", Ex.toString());    
            WriteEventLogEntry(sStr);
            return;
        }

        m_ConnectionArray[iIndex].bInUse = true;
        try 
        {   
            m_ConnectionArray[iIndex].ConnectSock.setSoTimeout(500); // 100ms timeout (non-blocking)
            m_ConnectionArray[iIndex].InStream = m_ConnectionArray[iIndex].ConnectSock.getInputStream();
            m_ConnectionArray[iIndex].OutStream = m_ConnectionArray[iIndex].ConnectSock.getOutputStream();
        }
        catch (Exception Ex)
        {
            WriteEventLogEntry(Ex.toString());
            return;
        }

        String sAddress = m_ConnectionArray[iIndex].ConnectSock.getRemoteSocketAddress().toString();
        sStr = String.format("Client %s Connected", sAddress);
        WriteEventLogEntry(sStr);
        CountActiveConnections(); // Updates control
    }

    int GetIndexOfUnusedConnection()
    {
        for(int i = 0; i < MAX_CONNECTIONS; i++)
        {	
            if(m_ConnectionArray[i].bInUse == false)
            {
                return i;
            }
        }
        return -1;	// Signal that all connections are in use
    }

    int CountActiveConnections()
    {
        int iCount = 0;
        for(int i = 0; i < MAX_CONNECTIONS; i++)
        {	
            if(m_ConnectionArray[i].bInUse == true)
            {
                iCount++;;
            }
        }
        jTextField_NumConnectedClients.setText(Integer.toString(iCount)); // update number of connections control
        return iCount;
    }

    int CountActiveGames()
    {
        int iCount = 0;
        for(int i = 0; i < MAX_GAMES; i++)
        {	
            if(m_GameArray[i].bInUse == true)
            {
                iCount++;;
            }
        }
        jTextField_NumActiveGames.setText(Integer.toString(iCount)); // update number of games control
        return iCount;
    }

    int FindIndexOfConnectionForAlias(String sAlias)
    {
        for(int i = 0; i < MAX_CONNECTIONS; i++)
        {	
            if(true == m_ConnectionArray[i].sAlias.equalsIgnoreCase(sAlias))
            {
                return i;
            }
        }
        return -1;	// Signal that Alias name was not found in connection array
    }

    int GetIndexOfUnusedGame()
    {
        for(int i = 0; i < MAX_GAMES; i++)
        {	
            if(m_GameArray[i].bInUse == false)
            {
                return i;
            }
        }
        return -1;	// Signal that all connections are in use
    }

    Boolean IsConnectionPartOfActiveGame(int iConnection)
    {
        for(int i = 0; i < MAX_GAMES; i++)
        {	
            if(m_GameArray[i].bInUse == true)
            {
                if(m_GameArray[i].iConnection[0] == iConnection	||
                        m_GameArray[i].iConnection[1] == iConnection)
                {	
                        return true;
                }
            }
        }
        return false;	// Signal connections is not part of an active game
    }

    int GetConnectionOpponent(int iConnection)
    {
        for(int i = 0; i < MAX_GAMES; i++)
        {	
            if(m_GameArray[i].bInUse == true)
            {
                if(m_GameArray[i].iConnection[0] == iConnection)
                {
                    return m_GameArray[i].iConnection[1];
                }	
                if(m_GameArray[i].iConnection[1] == iConnection)
                {
                    return m_GameArray[i].iConnection[0];
                }
            }
        }
        return -1;
    }

    void Start_Receive_Timer()
    {
        m_Receive_Timer = new Timer();
        m_Receive_Timer.scheduleAtFixedRate(
                new OnTimedEvent_Receive_Timer(), 100, 100); // Short interval to allow UI responsiveness)
    }

    void StopTimer()
    {
        if (null != m_Receive_Timer)
        {
            m_Receive_Timer.cancel();
        }
    }

    class OnTimedEvent_Receive_Timer extends TimerTask
    {
        public void run()
        {
            try
            {
                Accept();
                for (int i = 0; i < MAX_CONNECTIONS; i++)
                {
                    if (m_ConnectionArray[i].bInUse == true)
                    {
                        DoReceive(i);	// For each valid listen-socket DoReceive()
                    }
                }
            }
            catch (Exception Ex)
            {
                WriteEventLogEntry(Ex.toString());
            }
        }
    }
    
    void Start_SendServerAdvertisement_Timer()
    {
        m_SendServerAdvertisement_Timer = new Timer();
        m_SendServerAdvertisement_Timer.scheduleAtFixedRate(
                new OnTimedEvent_SendServerAdvertisement_Timer(), 1000, 1000); // 1 second interval
    }
    
    class OnTimedEvent_SendServerAdvertisement_Timer extends TimerTask
    {
        public void run()
        {
            Send_ServerAdvertisement_Broadcast();
        }
    }  

    void SendMessage(int iConnection, Game_Message_PDU Message)
    {
        byte[] bBuffer = Serialize_Game_Message_PDU(Message);
        try
        {
            m_ConnectionArray[iConnection].OutStream.write(bBuffer);
        }
        catch (SocketTimeoutException Ex)
        {
            return;
        }
        catch (Exception ex)
        {
            CloseGameAndConnection(iConnection);
            return;
        }
        String sStr = String.format("Sent:%s to:%s", Message.eMessageType.toString(), 
                m_ConnectionArray[iConnection].ConnectSock.getRemoteSocketAddress().toString());
        WriteEventLogEntry(sStr);
    }

    void SendPlayerList()
    {
        Game_Message_PDU Message = Message_PDU_Initialise(); 
        Message.eMessageType = MessageType.PLAYER_LIST;
        String sPlayerList = "";
        // List should include all Aliases that are connected, but not involved in active games
        for(int i = 0; i < MAX_CONNECTIONS; i++)
        {	
            if(m_ConnectionArray[i].bInUse == true && IsConnectionPartOfActiveGame(i) == false &&
                        m_ConnectionArray[i].sAlias.equalsIgnoreCase("") == false)
            {
                sPlayerList += m_ConnectionArray[i].sAlias;
                sPlayerList += ","; // Needs to be a comma-separated list
            }
        }
        if(sPlayerList.length() > 0)
        {
            sPlayerList = TruncateString(sPlayerList, sPlayerList.length() - 1); // Null terminate (write over last comma)
        }

        System.arraycopy(sPlayerList.getBytes()/*src*/, 0, Message.cPlayerList/*dest*/, 0, sPlayerList.length()/*count*/); 
        // Send PlayerList to ALL connected players
        for(int i = 0; i < MAX_CONNECTIONS; i++)
        {	
            if(m_ConnectionArray[i].bInUse == true)
            {
                SendMessage(i, Message);
            }
        }
    }

    void SendOpponentCellSelectionMessage(int iConnection, int iCell)
    {
        int iOpponentConnection;
        iOpponentConnection = GetConnectionOpponent(iConnection);   // Find players opponent
        if(iOpponentConnection < 0)
        {
            String sStr = String.format("Opponent not found for connection %d",iConnection);
            WriteEventLogEntry(sStr);
            return;
        }
        Game_Message_PDU Message = Message_PDU_Initialise(); 
        Message.eMessageType = MessageType.OPPONENT_CELL_SELECTION;
        Message.iCell = iCell;
        SendMessage(iOpponentConnection, Message);
    }

    void DoReceive(int iConnection)
    {		
        String sStr;
        byte[] bBuffer = new byte[1024];  // Create an appropriately-sized buffer to Receive into
        int iReceiveByteCount;
        try
        {
            iReceiveByteCount = m_ConnectionArray[iConnection].InStream.read(bBuffer);
        }
        catch (SocketTimeoutException Ex)
        {
            return;
        }
        catch (Exception Ex)
        {
            sStr = String.format("Receive failed, %s", Ex.toString());    
            WriteEventLogEntry(sStr);
            CloseConnection(iConnection);
            return;
        }
        if(-1 == iReceiveByteCount)
        {
            return; // no data
        }
        if(Game_Message_PDU_Size == iReceiveByteCount)
        {
            // Deserialise the Receive buffer to get the Game message structure
            Game_Message_PDU Message = DeSerialize_Game_Message_PDU(bBuffer);
            String sAlias = new String(Message.cAlias);
            sAlias = TruncateString(sAlias, MAX_ALIAS_LEN);
            int ilength = sAlias.length();
            switch (Message.eMessageType)
            {
                case REGISTER_ALIAS:
                    if (0 < ilength)
                    {
                        m_ConnectionArray[iConnection].sAlias = sAlias;
                        SendPlayerList();
                        sStr = String.format("Alias '%s' Registered", sAlias);
                        WriteEventLogEntry(sStr);
                    }
                    break;
                case CHOOSE_OPPONENT:
                    if(CreateGame(iConnection, Message) == false)
                    {
                        if (0 < ilength)
                        {
                            sStr = String.format("ChooseOpponent '%s' FAILED", sAlias);
                            WriteEventLogEntry(sStr);
                        }
                    }
                    break;
                case LOCAL_CELL_SELECTION:
                    SendOpponentCellSelectionMessage(iConnection, Message.iCell);
                    SetCellInGame(iConnection, Message.iCell);
                    if(!CheckForWin(iConnection))
                    {
                        EndOfGame(iConnection);	// Check for end of game (all cells full)
                    }
                    break;
                case CLIENT_DISCONNECT:
                    sStr = String.format("Client %s Disconnected", m_ConnectionArray[iConnection].sAlias);
                    WriteEventLogEntry(sStr);
                    CloseGameAndConnection(iConnection);
                    break;
            }
        }
    }
    
    Boolean CreateGame(int iConnection, Game_Message_PDU RecdMessage)
    {
        // Check initiating connection is not part of a game
        if(IsConnectionPartOfActiveGame(iConnection) == true)
        {
                return false;
        }
        // Check chosen opponent alias is connected
        String sAlias = new String(RecdMessage.cAlias);
        sAlias = TruncateString(sAlias, MAX_ALIAS_LEN);
        int ilength = sAlias.length();
        if (0 >= ilength)
        {
            return false;
        }

        int iOpponentConnection = FindIndexOfConnectionForAlias(sAlias);
        if(iOpponentConnection == -1)
        {
                return false;
        }
        // Check chosen opponent alias connection is not part of a game
        if(IsConnectionPartOfActiveGame(iOpponentConnection) == true)
        {
                return false;
        }
        // Get next available game structure in array
        int iGame = GetIndexOfUnusedGame();
        if(iGame == -1)
        {
                return false;
        }

        m_GameArray[iGame].iConnection[0] = iConnection;
        m_GameArray[iGame].iConnection[1] = iOpponentConnection;
        m_GameArray[iGame].bInUse = true;
        CountActiveGames();                 // Updates control
        Boolean bFlag = CreateRandomBool();    // Determine who goes first

        // Initialise game Grid
        for(int j = 0; j < 3; j++)
        {	
            for(int k = 0; k < 3; k++)
            {	
                m_GameArray[iGame].iGrid[j][k] = -1; // -1 = cell unoccupied (0 = '0' 1 = 'X')
            }	
        }

        SendPlayerList(); // Send updated list of available players

        // Send START_OF_GAME messages to both players
        Game_Message_PDU message = Message_PDU_Initialise(); 
        message.eMessageType = MessageType.START_OF_GAME;
        System.arraycopy(m_ConnectionArray[iOpponentConnection].sAlias.getBytes()/*src*/, 0, message.cAlias/*dest*/, 0, m_ConnectionArray[iOpponentConnection].sAlias.length()/*count*/); 
        message.iToken = GetPlayerToken(iGame, iConnection);	
        message.bFirstMove = bFlag;
        SendMessage(iConnection, message);

        System.arraycopy(m_ConnectionArray[iConnection].sAlias.getBytes()/*src*/, 0, message.cAlias/*dest*/, 0, m_ConnectionArray[iConnection].sAlias.length()/*count*/); 
        message.iToken = GetPlayerToken(iGame, iOpponentConnection);	
        message.bFirstMove = !bFlag;
        SendMessage(iOpponentConnection, message);

        String sStr = String.format("Game '%s-%s' created", 
                m_ConnectionArray[m_GameArray[iGame].iConnection[0]].sAlias,
                m_ConnectionArray[m_GameArray[iGame].iConnection[1]].sAlias);
        WriteEventLogEntry(sStr);
        return true;
    }

    int FindGameForConnection(int iConnection)			
    {
        for(int i = 0; i < MAX_GAMES; i++)
        {
        if (true == m_GameArray[i].bInUse && 
            (m_GameArray[i].iConnection[0] == iConnection ||
            m_GameArray[i].iConnection[1] == iConnection))
            {
                return i;
            }
        }
        return -1;
    }

    int GetPlayerToken(int iGame, int iConnection)			
    {
        // Lowest numbered connection will be given token '0'
        if(iGame >= 0 && iGame < MAX_GAMES)
        {
            if(	(m_GameArray[iGame].iConnection[0] == iConnection && 
                m_GameArray[iGame].iConnection[0] < m_GameArray[iGame].iConnection[1])
                    ||
                (m_GameArray[iGame].iConnection[1] == iConnection && 
                m_GameArray[iGame].iConnection[1] < m_GameArray[iGame].iConnection[0]) )
            {
                return 0;	// player identified by iConnection is '0'
            }
            else
            {
                return 1;	// player identified by iConnection is '1'
            }
        }
        return -1;	// iGame out of range
    }

    void DoSetCellInGame(int iGame, int iCell, int iToken)			
    {
        if((iToken == 0 || iToken == 1) && (iGame >= 0 && iGame < MAX_GAMES) 
                && (iCell >= 0 && iCell < 9))
        {
            // Cell 0 [0][0], 1 [0][1], 2 [0][2]
            //      3 [1][0], 4 [1][1], 5 [1][2]
            //      6 [2][0], 7 [2][1], 8 [2][2]
            m_GameArray[iGame].iGrid[iCell / 3][iCell % 3] = iToken;
        }
        else
        {
            String sStr = String.format("Could not set cell %d, in game %d", iCell, iGame);
            WriteEventLogEntry(sStr);
        }
    }

    void SetCellInGame(int iConnection, int iCell)			
    {
        String sStr = "";
        int iGame;
        iGame  = FindGameForConnection(iConnection);
        if(iGame  == -1)
        {	
            sStr = String.format("Game for connection %d not found", iConnection);
            WriteEventLogEntry(sStr);
            return;
        }

        // Check if player is '0' or 'X' ('0' is always assigned to the lowest numbered connection)
        int iToken;
        iToken = GetPlayerToken(iGame, iConnection);
        if(iToken  == -1)
        {	
            sStr = String.format("Token for connection %d (game %d) not found", iConnection, iGame);
            WriteEventLogEntry(sStr);
            return;
        }
        DoSetCellInGame(iGame, iCell, iToken);	
    }
	
    Boolean CheckForLine(int iGame, int iToken)
    {
        // check for win on diagonals
        if(m_GameArray[iGame].iGrid[1][1] == iToken &&
                ( (m_GameArray[iGame].iGrid[0][0] == iToken &&
                        m_GameArray[iGame].iGrid[2][2] == iToken)
                        ||
                  (m_GameArray[iGame].iGrid[0][2] == iToken &&
                        m_GameArray[iGame].iGrid[2][0] == iToken) ) )
        {
            return true;
        }
        // check for horizontal line win
        for(int j = 0; j < 3; j++)
        {
                if(m_GameArray[iGame].iGrid[j][0] == iToken &&
                        m_GameArray[iGame].iGrid[j][1] == iToken &&
                        m_GameArray[iGame].iGrid[j][2] == iToken)
            {
                return true;
            }
        }
        // check for vertical line win
        for(int j = 0; j < 3; j++)
        {
                if(m_GameArray[iGame].iGrid[0][j] == iToken &&
                        m_GameArray[iGame].iGrid[1][j] == iToken &&
                        m_GameArray[iGame].iGrid[2][j] == iToken)
            {
                return true;
            }
        }
        return false;
    }

    Boolean CheckForWin(int iConnection)
    {
        // Player iConnection has just moved, thus only that player could have won
        int iGame = FindGameForConnection(iConnection);
        int iWinningToken = GetPlayerToken(iGame, iConnection);
        // Seach for a line through the grid
        Boolean bResult;
        bResult = CheckForLine(iGame, iWinningToken);
        if(bResult == true)
        {
            Game_Message_PDU Message = Message_PDU_Initialise();
            Message.eMessageType = MessageType.END_OF_GAME;
            Message.eCode = CodeType.END_OF_GAME_WIN;
            SendMessage(iConnection, Message);
            int iOpponentConnection = GetConnectionOpponent(iConnection);
            Message.eCode = CodeType.END_OF_GAME_LOSE;
            SendMessage(iOpponentConnection, Message);
            RemoveGame(iGame);
        }
        return bResult;
    }

    Boolean EndOfGame(int iConnection)
    {
        // If no cells free, signal end of game (Draw) to both players
        int iGame;
        iGame = FindGameForConnection(iConnection);

        for(int j = 0; j < 3; j++)
        {	
            for(int k = 0; k < 3; k++)
            {	
                if(m_GameArray[iGame].iGrid[j][k] == -1) // -1 = cell unoccupied (0 = '0' 1 = 'X')
                {
                    return false;	// At least one unoccupied cell
                }
            }	
        }
        Game_Message_PDU Message = Message_PDU_Initialise();
        Message.eMessageType = MessageType.END_OF_GAME;
        Message.eCode = CodeType.END_OF_GAME_DRAW;
        SendMessage(iConnection, Message);
        int iOpponentConnection = GetConnectionOpponent(iConnection);
        SendMessage(iOpponentConnection, Message);
        RemoveGame(iGame);
        return true;
    }

    void RemoveGame(int iGame)
    {
        m_GameArray[iGame].bInUse = false;
        CountActiveGames();         // Updates control
        SendPlayerList();
    }
      
    void InformOpponent_GameEndedDueToOtherPlayerDisconnect(int iGame, int iDisconnectedClient_Connection)
    {
        // First, identify which of the two connections in the game relates to the disconnected client, and thus which relates to the opponent
        if(iDisconnectedClient_Connection == m_GameArray[iGame].iConnection[0])
        {
            CloseConnection(m_GameArray[iGame].iConnection[0]);
            m_ConnectionArray[m_GameArray[iGame].iConnection[0]].bInUse = false;

            // Send an END_OF_GAME (END_OF_GAME_CLIENT_DISCONNECT) message to the opponent
            Game_Message_PDU Message = Message_PDU_Initialise();
            Message.eMessageType = MessageType.END_OF_GAME;
            Message.eCode = CodeType.END_OF_GAME_CLIENT_DISCONNECT;
            SendMessage(m_GameArray[iGame].iConnection[1], Message);
            m_GameArray[iGame].bInUse = false;
            CountActiveGames(); // updates control
        }
        else if(iDisconnectedClient_Connection == m_GameArray[iGame].iConnection[1])
        {
            CloseConnection(m_GameArray[iGame].iConnection[1]);
            m_ConnectionArray[m_GameArray[iGame].iConnection[1]].bInUse = false;

            // Send an END_OF_GAME (END_OF_GAME_CLIENT_DISCONNECT) message to the opponent
            Game_Message_PDU Message = Message_PDU_Initialise();
            Message.eMessageType = MessageType.END_OF_GAME;
            Message.eCode = CodeType.END_OF_GAME_CLIENT_DISCONNECT;
            SendMessage(m_GameArray[iGame].iConnection[0], Message);
            m_GameArray[iGame].bInUse = false;
            CountActiveGames(); // updates control
        }
    }      

    void CloseGameAndConnection(int iConnection)
    {	
        int iGame = FindGameForConnection(iConnection);
        if(iGame != -1)
        {
            // Send a message informing the opponent that the game has ended due to the other player disconnecting
            InformOpponent_GameEndedDueToOtherPlayerDisconnect(iGame, iConnection);
            RemoveGame(iGame);
        }
        CloseConnection(iConnection);
    }

    void CloseConnection(int iConnection)
    {
        try
        {
             m_ConnectionArray[iConnection].ConnectSock.close();    // Close the socket
        }
        catch (Exception Ex)
        {
        }
        m_ConnectionArray[iConnection].bInUse = false;
        m_ConnectionArray[iConnection].sAlias = "";
        SendPlayerList();
        CountActiveConnections();
    }

    private void CloseAndExit()
    {   // Close all sockets and exit the application
        StopTimer();
        for (int i = 0; i < MAX_CONNECTIONS; i++)   // Close active connections
        {
            if (m_ConnectionArray[i].bInUse == true)
            {
                CloseConnection(i);
            }
        }
        try
        {
            m_ListenSocket.close();
        }
        catch (Exception Ex)
        {
        }
        System.exit(0);
    }
        
    private void WriteEventLogEntry(String sStr)
    {
        sStr += "\n";
        jTextArea_EventLog.append(sStr);       
    }

    Boolean CreateRandomBool()
    {
        int iRand = m_Random.nextInt(1000);     // gets int in range 0-999
        if(iRand >= 500)
        {
            return true;
        }
        return false;
    }

    Game_Message_PDU Message_PDU_Initialise()
    {
        Game_Message_PDU M = new Game_Message_PDU();
        M.eMessageType = MessageType.END_OF_GAME; // This field is always overwritten as used in every instance
        M.cAlias = new byte[MAX_ALIAS_LEN + 1];
        M.cPlayerList = new byte[10 * (MAX_ALIAS_LEN + 1)];
        M.iCell = 0;
        M.eCode = CodeType.END_OF_GAME_DRAW;
        M.iToken = 0;
        M.bFirstMove = false;
        return M;
    }
        
    static byte[] IntToByteArray(int iVal)
    {
        byte[] bArray = new byte[4];
        bArray[0] = (byte)(iVal & 0x000000ff);
        bArray[1] = (byte)((iVal & 0x0000ff00) >> 8);
        bArray[2] = (byte)((iVal & 0x00ff0000) >> 16);
        bArray[3] = (byte)((iVal & 0xff000000) >> 24);
        return bArray;
    }

    public int ByteArraySize4_ToInt(byte[] bVal)
    {
        if(4 != bVal.length)
        {
            return -1;  // Signal invalid byte array length
        }
        int iResult = bVal[0] & 0xFF;
        iResult += ((bVal[1] & 0xFF) << 8);
        iResult += ((bVal[2] & 0xFF) << 16);
        iResult += ((bVal[3] & 0xFF) << 24);            
        return iResult;
    }

    String TruncateString(String sStr, int iMaxLength)
    {
        // Truncate to null terminator if string contains null-termiation chars (happens when converting from a byte array)
        int iNullTerminatePosition = sStr.indexOf(0);
        if(-1 != iNullTerminatePosition)
        {
            sStr = sStr.substring(0, iNullTerminatePosition);
        }
       
        if(sStr.length()> iMaxLength)
        {   // Only use substring if the string is actually longer than the second parameter
            // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
            sStr = sStr.substring(0, iMaxLength);
        }
        return sStr;
    }

    void Initialise_ServerAdvertisement()
    {   // Create m_SendServerAdvertisementsSocket and set it NonBlocking and Broadcast
        try
        {
            m_SendServerAdvertisementSocket = new DatagramSocket(LOCAL_PORT); // Create socket and bind it a specified port (uses the local address)
            m_SendServerAdvertisementSocket.setSoTimeout(100); // Non-blocking with a timeout value of 100 milliseconds
            m_SendServerAdvertisementSocket.setBroadcast(true);
        }
        catch (SocketException Ex)  // Catch exceptions from the socket operations
        {
            WriteEventLogEntry(Ex.toString());
            return;
        }
        m_bSendServerAdvertisementBuffer = Initialise_ServerAdvertisement_PDU();
        Start_SendServerAdvertisement_Timer();
    }

    byte[] Initialise_ServerAdvertisement_PDU()
    {   // Intialise as separate step to sending, so only has to be done once
        ServerAdvertise_Message_PDU Message = new ServerAdvertise_Message_PDU();
        Message.eMessageType =  MessageType.SERVER_ADVERTISE;
        Message.cServerName = new byte[MAX_SERVER_NAME_LEN];
        String sServerName = TruncateString("Noughts_and_Crosses_SERVER", MAX_SERVER_NAME_LEN);
        System.arraycopy(sServerName.getBytes()/*src*/, 0, Message.cServerName/*dest*/, 0, sServerName.length()/*count*/); 
        byte[] bLocalAddress = m_LocalAddress.getAddress();
        Message.iIP_Address_Byte_1 = bLocalAddress[0];
        Message.iIP_Address_Byte_2 = bLocalAddress[1];
        Message.iIP_Address_Byte_3 = bLocalAddress[2];
        Message.iIP_Address_Byte_4 = bLocalAddress[3];
        Message.iPort = m_iPort;
        byte[] bSendServerAdvertisementBuffer = Serialize_ServerAdvertise_Message_PDU(Message);
        return bSendServerAdvertisementBuffer;
    }

    void Send_ServerAdvertisement_Broadcast()
    {
        try
        { 
            InetAddress BroadcastAddress = InetAddress.getByName("255.255.255.255");
            DatagramPacket sendPacket = new DatagramPacket(m_bSendServerAdvertisementBuffer,
                    ServerAdvertise_Message_PDU_Size, BroadcastAddress, SERVER_ADVERTISE_PORT); 
            m_SendServerAdvertisementSocket.send(sendPacket); 

            // Uncomment the following statement for diagnostic purposes
            // WriteEventLogEntry("Server Advertisement message broadcasted"); 
        } 
        catch (Exception Ex) 
        { 
            WriteEventLogEntry(Ex.toString()); 
        } 
    }
        
    private byte[] Serialize_Game_Message_PDU(Game_Message_PDU Message)
    {   // Convert an Message_PDU structure into a byte array
        byte[] bByteArray = new byte[Game_Message_PDU_Size];    // Create a byte array to hold the serialised form of the message PDU
        bByteArray[0] = (byte)Message.eMessageType.ordinal();
        System.arraycopy(Message.cAlias/*src*/, 0, bByteArray/*dest*/, 4, Message.cAlias.length/*count*/);
        System.arraycopy(Message.cPlayerList/*src*/, 0, bByteArray/*dest*/, 20, Message.cPlayerList.length/*count*/);
        byte[] bInt = IntToByteArray(Message.iCell);
        System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 180, 4/*count*/); // Next 4-byte boundary
        bByteArray[184] = (byte)Message.eCode.ordinal();
        bInt = IntToByteArray(Message.iToken);
        System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 188, 4/*count*/); // Next 4-byte boundary
    
        bByteArray[192] = (byte)0;  // Boolean serialises to 1 byte
        if(true == Message.bFirstMove)
        {   
            bByteArray[192] = (byte)1;
        }
        return bByteArray;
    }

    public Game_Message_PDU DeSerialize_Game_Message_PDU(byte[] bRecdByteArray)
    {   // Converts a byte array (from Receive) to a EventNotificationService_Message_PDU structure (Fields have fixed length, but non-fixed data length)
        Game_Message_PDU Game_PDU = new Game_Message_PDU();
        Game_PDU.eMessageType = MessageType.values()[bRecdByteArray[0]]; // Convert byte value to enumerated MessageType
        Game_PDU.cAlias = new byte[MAX_ALIAS_LEN+1];
        System.arraycopy(bRecdByteArray/*src*/, 4, Game_PDU.cAlias/*dest*/, 0, MAX_ALIAS_LEN/*count*/); 
        Game_PDU.cPlayerList = new byte[10 * (MAX_ALIAS_LEN+1)];
        System.arraycopy(bRecdByteArray/*src*/, 20, Game_PDU.cPlayerList/*dest*/, 0, (10 * (MAX_ALIAS_LEN+1))-1/*count*/); 
        byte[] bInt = new byte[4]; 
        System.arraycopy(bRecdByteArray/*src*/, 180, bInt/*dest*/, 0, 4/*count*/); 
        Game_PDU.iCell = ByteArraySize4_ToInt(bInt);            
        Game_PDU.eCode = CodeType.values()[bRecdByteArray[184]]; // Convert byte value to enumerated MessageType
        System.arraycopy(bRecdByteArray/*src*/, 188, bInt/*dest*/, 0, 4/*count*/); 
        Game_PDU.iToken = ByteArraySize4_ToInt(bInt); 
        Game_PDU.bFirstMove = true;
        if((byte)0 == bRecdByteArray[193])
        {
            Game_PDU.bFirstMove = false;
        }
        return Game_PDU;
    }

    private byte[] Serialize_ServerAdvertise_Message_PDU(ServerAdvertise_Message_PDU Message)
    {   // Convert an ServerAdvertise_Message_PDU structure into a byte array
        byte[] bByteArray = new byte[ServerAdvertise_Message_PDU_Size]; // Create a byte array to hold the serialised form of the message PDU
        bByteArray[0] = (byte)Message.eMessageType.ordinal(); // Encoded as a 4-byte field for compatibility with C# / C++ versions
        System.arraycopy(Message.cServerName/*src*/, 0, bByteArray/*dest*/, 4, Message.cServerName.length/*count*/);

        bByteArray[35] = Message.iIP_Address_Byte_1;
        bByteArray[36] = Message.iIP_Address_Byte_2;
        bByteArray[37] = Message.iIP_Address_Byte_3;
        bByteArray[38] = Message.iIP_Address_Byte_4;

        // Serialisation of port number (must start on a 4-byte boundary for compatibility with C# and C++ versions, so skip one byte
        // The message format and serialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
        // The port number is sent from the server as a 4-byte integer value of the format: high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
        // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received by clients, as:
        // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
        bByteArray[40] = (byte) (Message.iPort / 256); // Create the MSB byte of the port value
        bByteArray[41] = (byte) (Message.iPort % 256); // Create the LSB byte of the port value
        bByteArray[42] = 0; // Complete the encoding of the additional bytes (so that the natively Int16 port value is encoded as Int32)
        bByteArray[43] = 0; // Complete the encoding of the additional bytes (so that the natively Int16 port value is encoded as Int32)
        return bByteArray;
    }
    
    public static void main(String args[]) throws Exception
    {
        // Initialise the GUI libraries
        try 
        {
            javax.swing.UIManager.LookAndFeelInfo[] installedLookAndFeels=javax.swing.UIManager.getInstalledLookAndFeels();
            for (int idx=0; idx<installedLookAndFeels.length; idx++)
            {
                if ("Nimbus".equals(installedLookAndFeels[idx].getName())) {
                    javax.swing.UIManager.setLookAndFeel(installedLookAndFeels[idx].getClassName());
                    break;
                }
            }
        } 
        catch (Exception Ex)
        {
            System.exit(0);
        }

        // Create and display the GUI form
        java.awt.EventQueue.invokeAndWait(new Runnable()
        { 
            public void run() 
            { 
                new Game_Server_ServerAdvertise_ClientDisconnect().setVisible(true);
            }
        } );
    }    

    public void actionPerformed(ActionEvent e)
    {
        if(jButton_ClearLog == e.getSource())
        {   // Clear the Event Log text area
            jTextArea_EventLog.setText(null);
        }
        if(jButton_Done == e.getSource())
        {
            CloseAndExit();
        }
        if(jButton_EnableServer == e.getSource())
        {
            CreateSocket();
            jTextField_Port.setEnabled(false);
            jButton_EnableServer.setEnabled(false);
            Initialise_ServerAdvertisement();
            Start_Receive_Timer();   // All activity is non-blocking
        }
    }
        
    private javax.swing.JPanel jPanel_Communication;
    private javax.swing.JLabel jLabel_IPAddress;
    private javax.swing.JTextField jTextField_IPAddress;
    private javax.swing.JLabel jLabel_Port;
    private javax.swing.JTextField jTextField_Port;
    private javax.swing.JButton jButton_EnableServer;
    private javax.swing.JLabel jLabel_NumConnectedClients;
    private javax.swing.JTextField jTextField_NumConnectedClients;
    private javax.swing.JLabel jLabel_NumActiveGames;
    private javax.swing.JTextField jTextField_NumActiveGames;
    private javax.swing.JPanel jPanel_Controls;
    private javax.swing.JButton jButton_ClearLog;
    private javax.swing.JButton jButton_Done;
    private javax.swing.JPanel jPanel_EventLog;
    private javax.swing.JTextArea jTextArea_EventLog;
    private javax.swing.JScrollPane jScrollPane_EventLog;

    private void initComponents()
    {
        jLabel_IPAddress = new javax.swing.JLabel();
        jLabel_IPAddress.setText("IPv4 Address");
        jTextField_IPAddress = new javax.swing.JTextField();
        jTextField_IPAddress.setMaximumSize(new Dimension(100, 4));
        jLabel_Port = new javax.swing.JLabel();
        jLabel_Port.setText("Port");
        jTextField_Port = new javax.swing.JTextField();
        jTextField_Port.setMaximumSize(new Dimension(50, 4));
        jButton_EnableServer = new javax.swing.JButton();
        jButton_EnableServer.setText("Enable Server");
        jButton_EnableServer.setMaximumSize(new Dimension(120, 6));
        jButton_EnableServer.addActionListener(this);
        jLabel_NumConnectedClients = new javax.swing.JLabel();
        jLabel_NumConnectedClients.setText("Number of Connected Clients");
        jTextField_NumConnectedClients = new javax.swing.JTextField();
        jTextField_NumConnectedClients.setMaximumSize(new Dimension(30, 4));
        jLabel_NumActiveGames = new javax.swing.JLabel();
        jLabel_NumActiveGames.setText("Number of Active Games");
        jTextField_NumActiveGames = new javax.swing.JTextField();
        jTextField_NumActiveGames.setMaximumSize(new Dimension(30, 4));
        jPanel_Communication = new javax.swing.JPanel();
        jPanel_Communication.setPreferredSize(new Dimension(200, 400));
        jPanel_Communication.setBorder(javax.swing.BorderFactory.createTitledBorder("Communication configuration"));

        org.jdesktop.layout.GroupLayout jPanel1Layout = new org.jdesktop.layout.GroupLayout(jPanel_Communication);
        jPanel_Communication.setLayout(jPanel1Layout);
        jPanel1Layout.setHorizontalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jPanel1Layout.createSequentialGroup()
                .add(jLabel_IPAddress)
                .add(jTextField_IPAddress))
            .add(jPanel1Layout.createSequentialGroup()
                .add(jLabel_Port)
                .add(jTextField_Port))
            .add(jButton_EnableServer)
            .add(jPanel1Layout.createSequentialGroup()
                .add(jLabel_NumConnectedClients)
                .add(jTextField_NumConnectedClients))
            .add(jPanel1Layout.createSequentialGroup()
                .add(jLabel_NumActiveGames)
                .add(jTextField_NumActiveGames)));

        jPanel1Layout.setVerticalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jPanel1Layout.createSequentialGroup()
            .add(jPanel1Layout.createParallelGroup()
                .add(jLabel_IPAddress)
                .add(jTextField_IPAddress))            
            .addContainerGap(10, 10)
            .add(jPanel1Layout.createParallelGroup()
                .add(jLabel_Port)
                .add(jTextField_Port))
            .addContainerGap(20, 20)
            .add(jButton_EnableServer)
            .addContainerGap(20, 20)
            .add(jPanel1Layout.createParallelGroup()
                .add(jLabel_NumConnectedClients)
                .add(jTextField_NumConnectedClients))
            .addContainerGap(10, 10)
            .add(jPanel1Layout.createParallelGroup()
                .add(jLabel_NumActiveGames)
                .add(jTextField_NumActiveGames))));
       
        jPanel_Controls = new javax.swing.JPanel();
        jButton_ClearLog = new javax.swing.JButton();
        jButton_ClearLog.setText("Clear Event Log");
        jButton_ClearLog.setMaximumSize(new Dimension(30, 4));
        jButton_ClearLog.addActionListener(this);
        jButton_Done = new javax.swing.JButton();
        jButton_Done.setText("Shutdown Server");
        jButton_Done.setMaximumSize(new Dimension(30, 4));
        jButton_Done.addActionListener(this);
        jPanel_Controls.setPreferredSize(new Dimension(100, 100));
        jPanel_Controls.setBorder(javax.swing.BorderFactory.createTitledBorder("Controls"));

        org.jdesktop.layout.GroupLayout jPanel2Layout = new org.jdesktop.layout.GroupLayout(jPanel_Controls);
        jPanel_Controls.setLayout(jPanel2Layout);
        jPanel2Layout.setHorizontalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jButton_ClearLog)
            .add(jButton_Done));
        jPanel2Layout.setVerticalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jPanel2Layout.createSequentialGroup()
                .add(jButton_ClearLog)
                .addContainerGap(10, 10)
                .add(jButton_Done)));
        jPanel2Layout.linkSize(new java.awt.Component[] {jButton_ClearLog, jButton_Done});

        jPanel_EventLog = new javax.swing.JPanel();
        jTextArea_EventLog = new javax.swing.JTextArea();
        jScrollPane_EventLog = new javax.swing.JScrollPane(jTextArea_EventLog,
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, 
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        jPanel_EventLog.setPreferredSize(new Dimension(300, 400));
        jPanel_EventLog.setBorder(javax.swing.BorderFactory.createTitledBorder("Event Log (Diagnostic)"));
        
        org.jdesktop.layout.GroupLayout jPanel3Layout = new org.jdesktop.layout.GroupLayout(jPanel_EventLog);
        jPanel_EventLog.setLayout(jPanel3Layout);
        jPanel3Layout.setHorizontalGroup(
            jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(jScrollPane_EventLog));
        jPanel3Layout.setVerticalGroup(
            jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(jScrollPane_EventLog));        

        org.jdesktop.layout.GroupLayout layout = new org.jdesktop.layout.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        getContentPane().setPreferredSize(new Dimension(640, 400));
         
        layout.setHorizontalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(layout.createSequentialGroup()
            .add(layout.createParallelGroup()            
                .add(jPanel_Communication)
                .add(jPanel_Controls))
            .add(jPanel_EventLog)));

        layout.setVerticalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(layout.createSequentialGroup()
            .add(jPanel_Communication)
            .add(jPanel_Controls))
            .add(jPanel_EventLog));

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        setTitle("CaseStudy_Noughts_and_Crosses_Server");
        pack();
    }                                             
}